@@ -18,19 +18,7 @@
 #define LO_ENDIAN_H
 
 #include <sys/types.h>
-
-#ifdef _MSC_VER
-#ifndef UINTSDEFINED
-#define UINTSDEFINED
-#define int32_t __int32
-#define int64_t __int64
-#define uint32_t unsigned __int32
-#define uint64_t unsigned __int64
-#define uint8_t unsigned __int8
-#endif
-#else
 #include <stdint.h>
-#endif
 
 #if defined(WIN32) || defined(_MSC_VER)
 #include <winsock2.h>
@@ -121,7 +109,7 @@ static LO_INLINE uint64_t lo_swap64(uint64_t x)
 #ifdef LO_BIGENDIAN
 #undef LO_BIGENDIAN
 #endif
-#if 1 == 2
+#if 0 == 2
 #ifdef __BIG_ENDIAN__
 #define LO_BIGENDIAN 1
 #else
@@ -132,7 +120,7 @@ static LO_INLINE uint64_t lo_swap64(uint64_t x)
 #endif
 #endif
 #else
-#define LO_BIGENDIAN 1
+#define LO_BIGENDIAN 0
 #endif
 
 /* Host to OSC and OSC to Host conversion macros */
